#include <iostream>
#include <thread>
#include <vector>

using namespace std;

vector<int> bounds(int parts, int size) {
    vector<int> bnd;
    int delta = size / parts;
    int reminder = size % parts;
    int N1 = 0, N2 = 0;
    bnd.push_back(N1);
    for (int i = 0; i < parts; ++i) {
        N2 = N1 + delta;
        if (i == parts - 1)
            N2 += reminder;
        bnd.push_back(N2);
        N1 = N2;
    }
    return bnd;
}

void sumV (int *a, int *b, int *c, int limI, int limS, int tid) {
    // cout << "Sum " << tid << " " << limI << " " << limS << endl;
    for (int i = limI; i < limS; ++i) {
        c[i] = a[i] + b[i];
        // cout << "suma" << a[i] << "+" << b[i] << "=" << c[i] << endl;
    }
}


void serial(int* a, int* b, int* c, long int size) {
    sumV(ref(a), ref(b), ref(c), 0, size, 0);
}


void paralelo(int* a, int* b, int* c, long int size) {
    int parts = 8;
    vector<int> bnd = bounds(parts, size);
    thread *tt = new thread[parts];
    for (int i = 0; i < parts; ++i) {
        // cout << "Lanzando " << i << " " << bnd[i] << " " << bnd[i+1] << endl;
        tt[i] = thread(sumV, ref(a), ref(b), ref(c), bnd[i], bnd[i+1], i);
    }
    for (int i = 0; i < parts; ++i) {
        tt[i].join();
    }
    delete [] tt;
}


float _inner_measure(long int size, void(*func)(int*, int*, int*, long int)) {
    typedef chrono::high_resolution_clock Clock;
    Clock::time_point t0, t1;
    chrono::duration<float> diff;

    t0 = Clock::now();
    int *a = new int[size];
    int *b = new int[size];
    for (int i = 0; i < size; i++) {
        a[i] = 1;
        b[i] = 3;
    }
    int *c = new int[size];
    func(a, b, c, size);
    delete [] a;
    delete [] b;
    t1 = Clock::now();
    diff = t1 - t0;

    return diff.count();
}

 
float measure(long int size, void(*func)(int*, int*, int*, long int)) {
    // the function will be called 'inner_quant' times and averaged; that will happen
    // 'outer_quant' times, and taken the min of it (to simulate behaviour of .py measures)
    int inner_quant = 2; //20;
    int outer_quant = 3; //5;

    float result = 100000000;  // very big
    for (int i = 0; i < outer_quant; i++) {
        float inner_measure = 0;
        for (int j = 0; j < inner_quant; j++) {
            float inner_partial = _inner_measure(size, func);
            // cout << "Inner partial: " << inner_partial << endl;
            inner_measure += inner_partial;
        }
        inner_measure /= inner_quant;
        // cout << "Inner final: " << inner_measure << endl;
        if (inner_measure < result) {
            result = inner_measure;
        }
    // cout << "Result: " << result << endl;
    }
    
    return result;
}


void test(long int size) {
    typedef chrono::high_resolution_clock Clock;
    Clock::time_point t0, t1;
    chrono::duration<float> diff;

    cout << "Probando y midiendo para tamaño " << size << endl;

    cout << "  serial..." << endl;
    t0 = Clock::now();
    int *a1 = new int[size];
    int *b1 = new int[size];
    for (int i = 0; i < size; i++) {
        a1[i] = 1;
        b1[i] = 3;
    }
    int *c1 = new int[size];
    serial(a1, b1, c1, size);
    delete [] a1;
    delete [] b1;
    t1 = Clock::now();
      
    diff = t1 - t0;
    cout << "    " << diff.count() << endl;

    cout << "  paralelo..." << endl;
    t0 = Clock::now();
    int *a2 = new int[size]{1};
    int *b2 = new int[size]{3};
    for (int i = 0; i < size; i++) {
        a2[i] = 1;
        b2[i] = 3;
    }
    int *c2 = new int[size];
    paralelo(a2, b2, c2, size);
    delete [] a2;
    delete [] b2;
    t1 = Clock::now();

    diff = t1 - t0;
    cout << "    " << diff.count() << endl;

    cout << "Comparando" << endl;
    for (int i = 0; i < size; i++) {
        if (c1[i] != c2[i]) {
            cout << "Mal en pos " << i << endl;
            break;
        }
    }
    cout << "    ok" << endl;
    delete [] c1;
    delete [] c2;
}

int main() {
    // ejecuta ambas funciones y valida que den lo mismo
    test(50000);

    // mide para muchos tamaños
    long int sizes[] = {
        1000,
        2154,
        4641,
        10000,
        21544,
        46415,
        100000,
        215443,
        464158,
        1000000,
        2154434,
        4641588,
        10000000,
        21544346,
        46415888,
        100000000,
        0,
    };
    float result;
    int idx = 0;
    long int size;

    cout << "Midiendo..." << endl;
    while (1) {
        size = sizes[idx++];
        if (size == 0) {
            break;
        }
        result = measure(size, serial);
        cout << "serial;" << size << ";" << result << endl;
        result = measure(size, paralelo);
        cout << "paralelo;" << size << ";" << result << endl;
    }
    cout << "Leeeesto!" << endl;
    return 0;
}
